#######################################################################
# Copyright (c) 2019 Josef Spidlen, Ph.D., FlowJo, LLC
#
# License
# The software is distributed under the terms of the 
# Artistic License 2.0
# http://www.r-project.org/Licenses/Artistic-2.0
# 
# Disclaimer
# This software and documentation come with no warranties of any kind.
# This software is provided "as is" and any express or implied 
# warranties, including, but not limited to, the implied warranties of
# merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the  copyright holder be liable for any direct, 
# indirect, incidental, special, exemplary, or consequential damages
# (including but not limited to, procurement of substitute goods or 
# services; loss of use, data or profits; or business interruption)
# however caused and on any theory of liability, whether in contract,
# strict liability, or tort arising in any way out of the use of this 
# software.
######################################################################
######################################################################
## For debugging to see what FlowJo is passing to the script.
##
## cat("FJ-PAR-ADD-CELLIDS-TO-RESULT = FJ_PAR_ADD_CELLIDS_TO_RESULT \n")
## cat("FJ-DATA-FILE-PATH            = FJ_DATA_FILE_PATH \n")
## cat("FJ-OUTPUT-DIRECTORY-PATH     = FJ_OUTPUT_DIRECTORY_PATH \n")
## cat("FJ-RUNID                     = FJ_RUNID \n")
## cat("FJ-POPNAME                   = FJ_POPNAME \n")
## cat("FJ-PAR-DER-PAR-NAME          = FJ_DERIVED_PAR_NAME \n")
## cat("FJ-K-VALUE                   = FJ_K_VALUE \n")
######################################################################
######################################################################

stopAndReturnError <- function(message, ...) {
	cat(message, file=paste0(outPutFolder, "Error_Message_from_R.txt"))
	stop(message, ...)
}

#######################################################################
###### Start package import or installations
#######################################################################
# Install function for packages
loadPackages <-function(...) {
	libs<-unlist(list(...))
	req<-unlist(lapply(libs,require,character.only=TRUE))
	need<-libs[req==FALSE]
	if(length(need)>0){
		install.packages(need, repos="https://cran.r-project.org")
		lapply(need,require,character.only=TRUE)
	}
}

tryCatch({
  loadPackages(c('dplyr', 'viridisLite', 'magrittr', 'colourvalues', 'pheatmap', 'ggplot2', 'Rcpp', 'RANN', 'igraph', 'R6' ))
}, error = function(e) {
  message <- "There was an error with installing R packages. Please see How-To PDF and try to manually install."
  stopAndReturnError(message = message)
}, warning = function(w){
  loadPackages(c('dplyr', 'viridisLite', 'magrittr', 'colourvalues', 'pheatmap', 'ggplot2', 'Rcpp', 'RANN', 'igraph', 'R6'))
  writeLines("loaded R packages")
})

sessionInfo()

#######################################################################
#######################################################################
EventNumberDP <- "EventNumberDP"
CellIdDP <- "CellId"
addCellIdsToResult <- FJ_PAR_ADD_CELLIDS_TO_RESULT
parNames <- c(FJ_PARAMS_LIST)
fjDataFilePath <- "FJ_DATA_FILE_PATH"
outPutFolder <- "FJ_OUTPUT_DIRECTORY_PATH"
runID <- "FJ_RUNID"
fjPopname <- "FJ_POPNAME"
fjDerParName <- "FJ_DERIVED_PAR_NAME"
fjK <- FJ_K_VALUE

if (fjDerParName == "") {
	derParName <- paste0("Phenograph.", fjPopname, ".", runID)
} else {
	derParName <- paste0(fjDerParName, ".", runID)
}

## Returns a list of 3 elements
## X = a matrix of data from a CSV file as exported by FlowJo
## cellIds = a matrix with CellIds (if available)
## eventNumbers = a matrix with EventNumberDP (if available)
##
## csvPath: path the the CSV file as exported from FlowJo (including header with both stain and parameter is recommended when exporting from FlowJo)
readFlowJoCSV <- function(csvPath) {
	data <- read.csv(csvPath, check.names=FALSE)
	cellIdsColumn <- rep(NA, nrow(data))
	cellIds <- data.frame(CellIdDP = cellIdsColumn)
	names(cellIds) <- CellIdDP
	if (CellIdDP %in% names(data)) {
		cellIds <- data[, CellIdDP]
		data <- data[, -which(names(data) %in% c(CellIdDP))]
	}
	
	EventNumberColumn <- rep(NA, nrow(data))
	eventNumbers <- data.frame(EventNumberDP = EventNumberColumn)
	names(eventNumbers) <- EventNumberDP
	if (EventNumberDP %in% names(data)) {
		eventNumbers <- data[, EventNumberDP]
		data <- data[, -which(names(data) %in% c(EventNumberDP))]
	}
	
	# Trim the leading and tailing whitespaces
	colnames(data) <- gsub("^\\s+|\\s+$", "", colnames(data))
	list(X=as.matrix(data), cellIds=as.matrix(cellIds), eventNumbers=as.matrix(eventNumbers))	
}

tmpRet = readFlowJoCSV(fjDataFilePath);
inputMatrix <- tmpRet$X
cellIds <- tmpRet$cellIds
eventNumbers <- tmpRet$eventNumbers
colnames(eventNumbers) <- c(EventNumberDP)
rm(tmpRet)

## DEBUG printout
summary(inputMatrix)

## DEBUG printout
parNames

## Run Rphenograph with input and arguments specified by the plugin user
#######################
#######################

find_neighbors <- function(data, k){
  nearest <- nn2(data, data, k, searchtype = "standard")
  return(nearest[[1]])
}

  data = inputMatrix
  k = fjK

    if(is.data.frame(data))
        data <- as.matrix(data)

    if(!is.matrix(data))
        stop("Wrong input data, should be a data frame of matrix!")

    if(k<1){
        stop("k must be a positive integer!")
    }else if (k > nrow(data)-2){
        stop("k must be smaller than the total number of points!")
    }

    message("Run Rphenograph starts:","\n",
        "  -Input data of ", nrow(data)," rows and ", ncol(data), " columns","\n",
        "  -k is set to ", k)

    cat("  Finding nearest neighbors...")
    t1 <- system.time(neighborMatrix <- find_neighbors(data, k=k+1)[,-1])
    cat("DONE ~",t1[3],"s\n", " Compute jaccard coefficient between nearest-neighbor sets...")

        #################################
        #################################
        #################################
        cppFunction(' NumericMatrix jaccard_coeff(NumericMatrix idx) {
        int nrow = idx.nrow(), ncol = idx.ncol();
        NumericMatrix weights(nrow*ncol, 3);
        int r = 0;
        for (int i = 0; i < nrow; i++) {
            for (int j = 0; j < ncol; j++) {
                int k = idx(i,j)-1;
                NumericVector nodei = idx(i,_);
                NumericVector nodej = idx(k,_);
                int u = intersect(nodei, nodej).size();  // count intersection number
                if(u>0){
                    weights(r, 0) = i+1;
                    weights(r, 1) = k+1;
                    weights(r, 2) = u/(2.0*ncol - u)/2;  // symmetrize the graph
                    r++;
                }
            }
        }
            return weights;
                    }')
        #################################
        #################################
        #################################

    t2 <- system.time(links <- jaccard_coeff(neighborMatrix))

    cat("DONE ~",t2[3],"s\n", " Build undirected graph from the weighted links...")
    links <- links[links[,1]>0, ]
    relations <- as.data.frame(links)
    colnames(relations)<- c("from","to","weight")
    t3 <- system.time(g <- graph.data.frame(relations, directed=FALSE))

    # Other community detection algorithms:
    #    cluster_walktrap, cluster_spinglass,
    #    cluster_leading_eigen, cluster_edge_betweenness,
    #    cluster_fast_greedy, cluster_label_prop
    cat("DONE ~",t3[3],"s\n", " Run louvain clustering on the graph ...")
    t4 <- system.time(community <- cluster_louvain(g))
    cat("DONE ~",t4[3],"s\n")

    message("Run Rphenograph DONE, totally takes ", sum(c(t1[3],t2[3],t3[3],t4[3])), "s.")
    cat("  Return a community class\n  -Modularity value:", modularity(community),"\n")
    cat("  -Number of clusters:", length(unique(membership(community))))

    Rphenograph_out <- list(g, community)

#######################
#######################

outputMatrix <- as.matrix(membership(Rphenograph_out[[2]])) 

if ((NA %in% outputMatrix) || (NaN %in% outputMatrix)) {
  stopAndReturnError("Unfortunately, we were unable to calculate results for this input\n(NaN in the resulting matrix). You may want to change the input arguments.")
} 

## Let's also save modularity
cat(modularity(Rphenograph_out[[2]]), file=paste0(outPutFolder, "modularity.txt"))

## Let's also save the name of the derived parameter
cat(derParName, file=paste0(outPutFolder, "parName.txt"))

## Let's also save the number of clusters
cat(max(outputMatrix), file=paste0(outPutFolder, "nClust.txt"))

if (addCellIdsToResult && !(NA %in% cellIds)) {

    ## SeqGeq

  colnames(outputMatrix) <- derParName

  tryCatch({
    phenographClustMatrix <- as.data.frame(cbind(outputMatrix, inputMatrix))
    summaryPhenoMatMedian <- phenographClustMatrix %>% dplyr::group_by_at(derParName) %>% summarise_all(median)
    rm(phenographClustMatrix)
    parsToPlot <- colnames(inputMatrix)
    summaryPhenoMatMedian <- dplyr::select(summaryPhenoMatMedian, dplyr::all_of(parsToPlot))

    # Get the number of rows and columns (parameters)
    num_rows <- nrow(summaryPhenoMatMedian)
    num_cols <- ncol(summaryPhenoMatMedian)

    # Set max and min font sizes
    max_font_size <- 8
    min_font_size <- 2

    # Adjust the constants to make scaling smoother
    row_font_size <- min(max_font_size, max(min_font_size, 50 / sqrt(num_rows)))
    col_font_size <- min(max_font_size, max(min_font_size, 20 / sqrt(num_cols)))

    # Print the selected font sizes for debugging purposes
    cat("Row font size:", row_font_size, "\n")
    cat("Column font size:", col_font_size, "\n")

    breaklist = seq(0, 4096, by=0.1)
    pheatmap::pheatmap(summaryPhenoMatMedian, scale = "none", fontsize_row = row_font_size,
                       fontsize_col = col_font_size, angle_col=90,
                       color = colourvalues::colour_values(breaklist, palette = "cividis"),
                       main=paste0("Phenograph Clusters of ", fjPopname), cluster_rows=TRUE, width=5, height=5,
                       file=paste0(outPutFolder, "Phenograph_ClusterHeatMap_", runID, ".png"),
                       bg="transparent", useDingbats=FALSE)

  }, error = function(e) {
    print("Failed to create heatmap.")
    message(conditionMessage(e))
  })


  # Prepare Results file
  outputMatrix <- cbind(cellIds, outputMatrix)
  colnames(outputMatrix) <- c(CellIdDP, derParName)
  write.table(outputMatrix, file=paste0(outPutFolder, "Results_Phenograph_", runID, ".csv"), row.names=FALSE, quote=FALSE, append=TRUE, sep=",")
} else {
    ## FlowJo
    colnames(outputMatrix) <- derParName

  tryCatch({
    # Generate Heatmap
    phenographClustMatrix <- as.data.frame(cbind(outputMatrix, inputMatrix))
    summaryPhenoMatMedian <- phenographClustMatrix %>% dplyr::group_by_at(derParName) %>% summarise_all(median)
    rm(phenographClustMatrix)
    parsToPlot <- colnames(inputMatrix)
    summaryPhenoMatMedian <- dplyr::select(summaryPhenoMatMedian, dplyr::all_of(parsToPlot))

    # Get the number of rows and columns (parameters)
    num_rows <- nrow(summaryPhenoMatMedian)
    num_cols <- ncol(summaryPhenoMatMedian)

    # Set max and min font sizes
    max_font_size <- 8
    min_font_size <- 3

    # Adjust the constants to make scaling smoother
    row_font_size <- min(max_font_size, max(min_font_size, 50 / sqrt(num_rows)))
    col_font_size <- min(max_font_size, max(min_font_size, 20 / sqrt(num_cols)))

    # Print the selected font sizes for debugging purposes
    cat("Row font size:", row_font_size, "\n")
    cat("Column font size:", col_font_size, "\n")

    breaklist = seq(0, 1023, by=0.1)
    pheatmap::pheatmap(summaryPhenoMatMedian, scale = "none", fontsize_row = row_font_size,
                       fontsize_col = col_font_size, angle_col=45,
                       color = colourvalues::colour_values(breaklist, palette = "inferno"),
                       main=paste0("Phenograph Clusters of ", fjPopname), cluster_rows=TRUE, width=5, height=5,
                       file=paste0(outPutFolder, "Phenograph_ClusterHeatMap_", runID, ".png"),
                       bg="transparent", useDingbats=FALSE)

  }, error = function(e) {
    print("Failed to create heatmap.")
    message(conditionMessage(e))
  })

  write.csv(cbind(eventNumbers, outputMatrix), file=paste0(outPutFolder, "Results_Phenograph_", runID, ".csv"), row.names=FALSE, quote=FALSE)
}

# R seems to be saving the .RData when exiting, so let's clean up to at least make that tiny (i.e., empty environment)
# rm(list=ls())
